#include <math.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "boxes.h"
#include "display.h"
#include "opts.h"
#include "needle_common.h"
#include "needlep_common.h"
#include "output.h"
#include "input.h"





extern int
box_length(const sequence* seqs, const int nb_seqs, sequence** prots);



struct blackbox_t blackboxes[] = {
   { "needle on nucleic sequences", "needle", "needle.out", "", "", box_needle},
   { "length", "length", "length.out", "", "", box_length},
   { "needle on amino acid sequences", "needlep", "needlep.out", "needlep_graph.tex", "needlep_ranked_graph.tex", box_needlep},
};

blackbox NEEDLE_IDENTITY_BOX = &(blackboxes[0]);
blackbox LENGTH_BOX = &(blackboxes[1]);
blackbox NEEDLE_PROTEIN_BOX = &(blackboxes[2]);




int
box_length(const sequence* seqs, const int nb_seqs, sequence** prots) {
   int i;
   double r = 0.0;
   FILE* output_fd = NULL;
   char* output = NULL;

   if(OPTS_output_dir) {
      NEW(output, char, (strlen(OPTS_output_dir) + strlen(LENGTH_BOX->out_file) + 1));
      strcpy(output, OPTS_output_dir);
      strcat(output, LENGTH_BOX->out_file);
      
      if((output_fd = fopen(output, "w")) == NULL) {
	 ERROR_ "unable to create output file '%s'\n", output _ERROR;
	 return 1;
      }
   } else
      output_fd = stdout;

   for(i = 0; i < nb_seqs; i++)
      r += seqs[i]->length;
      
   r /= nb_seqs;

   fprintf(output_fd, "length = %.0f\n", r);

   if(OPTS_output_dir) {
      fflush(output_fd);
      fclose(output_fd);
      DESTROY(output);
   }


   return 0;
}
