#ifdef MY_NEEDLE

#include "opts.h"
#include "display.h"
#include "needlep_common.h"
#include "boxes.h"
#include "output.h"
#include "my_needlep_matrices.h"

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <math.h>
#include <limits.h>







int
call_needlep(const sequence si, const sequence sj, float* res,  const float gapopen, const float gapextend, const float identity) {
   int status, i, j, mi, mj;
   float n1, n2, n3;
   float **prog_dyn_mm = NULL;
   float **prog_dyn_ins = NULL;
   float **prog_dyn_del = NULL;
   float *swaln = NULL;
   char *output = NULL;
   FILE *fd;
   char *matname = NULL;
   
   blosum_matrix subsmatrix;

   int maxi, maxj;

#ifndef GLOBAL_NEEDLE
   float max = -INFINITY;
   
   maxi = si->length;
   maxj = sj->length;
#else
   float temp;
#endif



 
   if(init_needlep()) return -1;

   i = 0;
   if((OPTS_output_dir != NULL) && (OPTS_keep_files != 0)) {
      NEW(output, char, (1 + strlen(OPTS_output_dir) + strlen(NEEDLE_PROTEIN_BOX->dir) + 1 + strlen(si->name) + (2 * strlen(STRAND_MARK)) + 8 + strlen(SEQUENCE_SEP) + strlen(sj->name)));
      
      sprintf(output, "%s%s/%s%s%+i%s%s%s%+i", OPTS_output_dir, NEEDLE_PROTEIN_BOX->dir, si->name, STRAND_MARK, si->strand, SEQUENCE_SEP, sj->name, STRAND_MARK, sj->strand);

      if(OPTS_use_cache) {
	 if((fd = fopen(output, "r")) != NULL) {

	    i = 1;
	    status = needle_protein_result(fd, res);
	    fclose(fd);
	    DESTROY(output);
	 }
      } else {
	 INFO_ "." _INFO;
      }
   }
   
   if(i == 0) {
      if(OPTS_produce_pairwise_alignements) {
	 WARN_ "BACKTRACKING IS NOT IMPLEMENTED IN THIS VERSION\n" _WARN;
      }

      /* initialisation */
      NEW(prog_dyn_mm, float*, 2);
      NEW(prog_dyn_ins, float*, 2);
      NEW(prog_dyn_del, float*, 2);
      
      
      for(j = 0; j < 2; j++) {
	 NEW(prog_dyn_mm[j], float, (sj->length + 1));
	 NEW(prog_dyn_ins[j], float, (sj->length + 1));
	 NEW(prog_dyn_del[j], float, (sj->length + 1)); 
      }
      
      
      
      for(j = 0; j <= sj->length; j++) {
	 prog_dyn_mm[0][j] = -INFINITY;
#ifdef GLOBAL_NEEDLE
	 prog_dyn_ins[0][j] = - gapopen - (gapextend * j);
#else
	 prog_dyn_ins[0][j] = 0.0;
#endif
	 prog_dyn_del[0][j] = -INFINITY;
      }

      prog_dyn_mm[0][0] = 0.0;
      prog_dyn_ins[0][0] = prog_dyn_del[0][0] = -INFINITY;
      
   
#ifdef ADAPTIVE_BLOSUM

      if(identity >= 90.0) {
	 subsmatrix = &blosum90;
	 matname = "BLOSUM90";
      } else if(identity >= 85.0) {
	 subsmatrix = &blosum85;
	 matname = "BLOSUM85";
      } else if(identity >= 80.0) {
	 subsmatrix = &blosum80;
	 matname = "BLOSUM80";
      } else if(identity >= 75.0) {
	 subsmatrix = &blosum75;
	 matname = "BLOSUM75";
      } else if(identity >= 70.0) {
	 subsmatrix = &blosum70;
	 matname = "BLOSUM70";
      } else if(identity >= 65.0) {
	 subsmatrix = &blosum65;
	 matname = "BLOSUM65";
      } else if(identity >= 62.0) {
	 subsmatrix = &blosum62;
	 matname = "BLOSUM62";
      } else if(identity >= 60.0) {
	 subsmatrix = &blosum60;
	 matname = "BLOSUM60";
      } else if(identity >= 55.0) {
	 subsmatrix = &blosum55;
	 matname = "BLOSUM55";
      } else if(identity >= 50.0) {
	 subsmatrix = &blosum50;
	 matname = "BLOSUM50";
      } else if(identity >= 45.0) {
	 subsmatrix = &blosum45;
	 matname = "BLOSUM45";
      } else if(identity >= 40.0) {
	 subsmatrix = &blosum40;
	 matname = "BLOSUM40";
      } else if(identity >= 35.0) {
	 subsmatrix = &blosum35;
	 matname = "BLOSUM35";
      } else if(identity >= 30.0) {
	 subsmatrix = &blosum30;
	 matname = "BLOSUM30";
      } else {
	 subsmatrix = &blosumn;
	 matname = "BLOSUMN";
      }

#else
      subsmatrix = &blosum62;
      matname = "BLOSUM62";
#endif
   
      /* boucle */
   
      for(i=1; i <= si->length; i++) {

	 prog_dyn_mm[1][0] = -INFINITY;
	 prog_dyn_ins[1][0] = -INFINITY;
#ifdef GLOBAL_NEEDLE
	 prog_dyn_del[1][0] = - gapopen - (gapextend * i);
#else
	 prog_dyn_del[1][0] = 0.0;
#endif
	 



	 
	 GET_AA_INDICE(si->bases[i-1], mi);
	 if(mi == -1) {
	    ERROR_ "SEQUENCE '%s' contains invalid base at position %i!!!\n", si->name, i _ERROR;
	    return -1;
	 }
	 
	 for(j=1; j <= sj->length; j++) {
	    
	    GET_AA_INDICE(sj->bases[j-1], mj);
	    if(mj == -1) {
	       ERROR_ "SEQUENCE '%s' contains invalid base at position %i!!!\n", sj->name, j _ERROR;
	       return -1;
	    }

	    prog_dyn_mm[1][j] = (*subsmatrix)[mi][mj] + MAX3(prog_dyn_mm[0][j-1], prog_dyn_ins[0][j-1], prog_dyn_del[0][j-1]);

	    n1 = - gapopen /*- gapextend*/ + prog_dyn_mm[1][j-1];
	    n2 = - gapextend + prog_dyn_ins[1][j-1];
	    n3 = - gapopen /*- gapextend*/ + prog_dyn_del[1][j-1];
	 
	    prog_dyn_ins[1][j] = MAX3(n1, n2, n3);

	    n1 = - gapopen /*- gapextend*/ + prog_dyn_mm[0][j];
	    n2 = - gapopen /*- gapextend*/ + prog_dyn_ins[0][j];
	    n3 = - gapextend + prog_dyn_del[0][j];

	    prog_dyn_del[1][j] = MAX3(n1, n2, n3);
	 }

#ifndef GLOBAL_NEEDLE
	 /*      max = MAX2(max, prog_dyn_mm[i][sj->length]);*/
	 /*max = MAX2(max, prog_dyn_mm[i][sj->length]); */
	 if(max < prog_dyn_mm[1][sj->length]) {
	    max = prog_dyn_mm[1][sj->length];
	    maxi = i;
	    maxj = sj->length;
	 }
#endif

	 swaln = prog_dyn_mm[0];
	 prog_dyn_mm[0] = prog_dyn_mm[1];
	 prog_dyn_mm[1] = swaln;

	 swaln = prog_dyn_del[0];
	 prog_dyn_del[0] = prog_dyn_del[1];
	 prog_dyn_del[1] = swaln;

	 swaln = prog_dyn_ins[0];
	 prog_dyn_ins[0] = prog_dyn_ins[1];
	 prog_dyn_ins[1] = swaln;
      }
      
      /* resultat */
   
#ifdef GLOBAL_NEEDLE
      if(prog_dyn_mm[0][sj->length] >= prog_dyn_ins[0][sj->length]) {
	 temp = prog_dyn_mm[0][sj->length];
      } else {
	 temp = prog_dyn_ins[0][sj->length];
      }
      if(prog_dyn_del[0][sj->length] > temp) {
	 temp = prog_dyn_del[0][sj->length];
      }
      
      (*res) = temp;

      maxi = i = si->length;
      maxj = j = sj->length;
#else


      /*    for(i = 1; i <= sj->length; i++) { */
      /*       max = MAX2(max, prog_dyn_mm[si->length][i]); */
      /*    } */
      for(j = 1; j <= sj->length; j++) {
	 if(max < prog_dyn_mm[0][j]) {
	    max = prog_dyn_mm[0][j];
	    maxi = si->length;
	    maxj = j;
	 }
      }

      i = maxi;
      j = maxj;


      (*res) = max;

      status = 0;
#endif



    
   
      /* nettoyage */

      for(j = 0; j < 2 ; j++) {
	 DESTROY(prog_dyn_mm[j]);
	 DESTROY(prog_dyn_ins[j]);
	 DESTROY(prog_dyn_del[j]);
      }

      DESTROY(prog_dyn_mm);
      DESTROY(prog_dyn_ins);
      DESTROY(prog_dyn_del);
      
      
      if(output) {
	 status = 0;
	 if((fd = fopen(output, "w")) != NULL) {
	    fprintf(fd, "# Score:\t %.2f\n", (*res));
	    fprintf(fd, "# Matrix:\t %s\n", matname);
	    fflush(fd);
	    fclose(fd);
	 } else {
	    ERROR_ "unable to create file '%s'\n", output _ERROR;
	    status = -1;
	 }
	 
	 DESTROY(output);
      }
   }

   if(status != -1) {
      if((*res) < OPTS_min_aln_score) {
	 status = 1;
      } else {
	 status = 0;
      }
      /*   status = 1;*/ /* seulement si resultat insatisfaisant */
      /*status = -1;*/ /* pb grave */
   }

   return status;
}






#endif
