/*
Copyright or © or CCopr. INRIA : Arnaud FONTAINE

arnaud.fontaine@lifl.fr

This software is a computer program whose purpose is to predict 
significantly conserved protein coding sequences and/or RNA structure 
on a set of unaligned nucleic sequences.

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#ifndef MY_NEEDLE

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>


#include "needlep_common.h"
#include "opts.h"
#include "display.h"
#include "boxes.h"
#include "input.h"
#include "output.h"



int
call_needlep(const sequence si, const sequence sj, float* res, const float gapopen, const float gapextend, const float identity) {
   pid_t pid;
   char* output;
   char* args[30];
   int status, cur_args;
   char *to_free_1 = NULL;
   char *to_free_2 = NULL;
   char *sw = NULL;
   FILE *fd = NULL;

   if(init_needlep()) return -1;

   if((OPTS_output_dir != NULL) && (OPTS_keep_files != 0)) {
      NEW(output, char, (1 + strlen(OPTS_output_dir) + strlen(NEEDLE_PROTEIN_BOX->dir) + 1 + strlen(si->name) + (2 * strlen(STRAND_MARK)) + 8 + strlen(SEQUENCE_SEP) + strlen(sj->name)));
      
      sprintf(output, "%s%s/%s%s%+i%s%s%s%+i", OPTS_output_dir, NEEDLE_PROTEIN_BOX->dir, si->name, STRAND_MARK, si->strand, SEQUENCE_SEP, sj->name, STRAND_MARK, sj->strand);
   } else {
      /*      output = tmpnam(NULL);*/
      sw = tmpnam(NULL);
      NEW(output, char, (1 + strlen(sw)));
      strcpy(output, sw);
   }


   if(((fd = fopen(output, "r")) == NULL) || (OPTS_use_cache == 0)) {

      INFO_ "." _INFO;
      if(fd) { fclose(fd); }

      cur_args = 0;
      args[cur_args++] = NEEDLE_EXE;
      args[cur_args++] = "-auto";
      /*args[cur_args++] = "-brief";*/
      args[cur_args++] = "-outfile";
      args[cur_args++] = output;
      args[cur_args++] = "-sprotein1";
      args[cur_args++] = "-sprotein2";
      args[cur_args++] = "-gapopen";
      asprintf(&(args[cur_args]), "%.2f", gapopen);
      to_free_1 = args[cur_args++];
      /*   args[cur_args++] = "7.0";*/
      args[cur_args++] = "-gapextend";
      /*   args[cur_args++] = "1.0";*/
      asprintf(&(args[cur_args]), "%.2f", gapextend);
      to_free_2 = args[cur_args++];
      args[cur_args++] = si->file; /*INFO_ "SEQI = %s\n", seqi->file _INFO;*/
      args[cur_args++] = sj->file; /*INFO_ "SEQJ = %s\n", seqj->file _INFO;*/
      /*args[cur_args++] = "-die";
	args[cur_args++] = "-warning";
	args[cur_args++] = "-error";
	args[cur_args++] = "-fatal";*/
      
      args[cur_args++] = "-datafile";


#ifdef ADAPTIVE_BLOSUM
      if(identity >= 90.0) args[cur_args++] = "EBLOSUM90";
      else if(identity >= 85.0) args[cur_args++] = "EBLOSUM85";
      else if(identity >= 80.0) args[cur_args++] = "EBLOSUM80";
      else if(identity >= 75.0) args[cur_args++] = "EBLOSUM75";
      else if(identity >= 70.0) args[cur_args++] = "EBLOSUM70";
      else if(identity >= 65.0) args[cur_args++] = "EBLOSUM65";
      else if(identity >= 62.0) args[cur_args++] = "EBLOSUM62";
      else if(identity >= 60.0) args[cur_args++] = "EBLOSUM60";
      else if(identity >= 55.0) args[cur_args++] = "EBLOSUM55";
      else if(identity >= 50.0) args[cur_args++] = "EBLOSUM50";
      else if(identity >= 45.0) args[cur_args++] = "EBLOSUM45";
      else if(identity >= 40.0) args[cur_args++] = "EBLOSUM40";
      else if(identity >= 35.0) args[cur_args++] = "EBLOSUM35";
      else if(identity >= 30.0) args[cur_args++] = "EBLOSUM30";
      else args[cur_args++] = "EBLOSUMN";
#else
      args[cur_args++] = "EBLOSUM62";
#endif
      args[cur_args++] = NULL;
      


      status = 0;
      if((pid = fork())) {
	 /* pere */
	 waitpid(pid, &status, 0);
      } else {
	 /* fils */
	 execvp(args[0], args);
	 exit(1);
      }
      
      if(WIFEXITED(status)) {
	 if(WEXITSTATUS(status) > 0) {
	    ERROR_ "'%s' has crashed or has not been found in from tour PATH !\n", args[0] _ERROR;
	    return -1;
	 }
      }
      
      if(to_free_1) {
	 DESTROY(to_free_1);
      }
      
      if(to_free_2) {
	 DESTROY(to_free_2);
      }
   }

   if(fd == NULL) {
      fd = fopen(output, "r");
   }

   status = needle_protein_result(fd, res);
   
   fclose(fd);

   if(status > 0) {
      /* pas d'echec, mais absence de rsultat */
      (*res) = 0.0;
   } else if(status == 0) {
	 /* pas d'chec, rsultat existe mais insatisfaisant */
      if((*res) < OPTS_min_aln_score) {
	 status = 1;
      }
   } else {
      /* probleme grave */
      ERROR_ "an error occured during %s result analysis\n", NEEDLE_PROTEIN_BOX->name _ERROR;
   }

   if((OPTS_output_dir == NULL) || (OPTS_keep_files == 0)) {
      unlink(output);
   }
   DESTROY(output);   

   return status;
}

#endif
