#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <regex.h>
#include <mpfr.h>


#include "results.h"
#include "reduce.h"
#include "pvalues.h"
#include "alignment.h"
#include "fusion.h"

#include "opts.h"

#include "opts.h"
#include "display.h"
#include "needlep_common.h"
#include "boxes.h"
#include "output.h"
#include "input.h"



regex_t* needlep_protein_regex2 = NULL;
regex_t* needlep_protein_score_regex = NULL;


int
init_needlep() {
   if(needlep_protein_regex2 == NULL) {
      NEW(needlep_protein_regex2, regex_t, 1);
      NEW(needlep_protein_score_regex, regex_t, 1);
       
      if(regcomp(needlep_protein_regex2, NEEDLEP_PROTEIN_REGEX2, REG_EXTENDED | REG_NEWLINE)) {
	 ERROR_ "REGULAR EXPRESSION needlep_protein_regex2 IN NEEDLE PROBLEM\n" _ERROR;
	 return 1;
      }
      
      if(regcomp(needlep_protein_score_regex, NEEDLEP_PROTEIN_SCORE_REGEX, REG_EXTENDED)) {
	 ERROR_ "REGULAR EXPRESSION needlep_protein_score_regex IN NEEDLE PROBLEM\n" _ERROR;
	 return 1;
      }
   }

   return 0;
}



int
needle_protein_result(/*const char* output*/FILE *fd, float* score) {
   int nb;
   char *line = NULL;
   char *trans = NULL, *trans2 = NULL;
   /* FILE* fd;*/
   regmatch_t regmatch;

   /*if((fd = fopen(output, "r"))) {*/
            
      NEW(trans, char, 100);
      NEW(trans2, char, 100);

      *score = -INFINITY; /* a negative value to force a first value */

      while(((*score < 0.0)) && (line = read_line(fd, &nb))) {
	 RENEW(trans, char, (nb + 1));
	 strncpy(trans, line, nb);
	 trans[nb] = '\0';
	 
	 if (regexec(needlep_protein_regex2, trans, 1, &regmatch, 0) == 0) {
	    RENEW(trans2, char, (nb + 1));
	    
	    regexec(needlep_protein_score_regex, trans, 1, &regmatch, 0);
	    strncpy(trans2, &(trans[regmatch.rm_so]), regmatch.rm_eo - regmatch.rm_so);
	    trans2[regmatch.rm_eo - regmatch.rm_so] = '\0';

	    *score = atof(trans2);
	 }
	 DESTROY(line);
      }
    /*  
      fclose(fd);
*/
      DESTROY(trans);
      DESTROY(trans2);

      if(*score == -INFINITY) return -1;
   /*} else {
      ERROR_ "unable to open %s output file '%s'\n", NEEDLE_PROTEIN_BOX->name, output _ERROR;
      return -1;
   }*/

   return 0;
}





float
****build_rfg(const sequence* seqs, int nb_seqs, sequence** prots, const float gapopen, const float gapextend, float* min, float* max) {
   int i, j, k, l, m, n, c;
   
   float ****results = NULL;


   INFO_ "Building RFG (%s)...", NEEDLE_PROTEIN_BOX->name _INFO;
   FLUSH_OUT;

   NEW(results, float***, (nb_seqs - 1));

   (*min) = +INFINITY;
   (*max) = -INFINITY;


   /* sequence 1 */
   for(i = 0; i < (nb_seqs - 1); i++) {
      if(OPTS_first_phase < 0) {
	 NEW(results[i], float**, 6);
      } else {
	 NEW(results[i], float**, 3);
      }

      /* phase sequence 1 */
      for(j = OPTS_first_phase; j < 4; j++) {
	 if(j != 0) {

	    m = (j > 0) ? (j - 1) : (2 - j);

	    NEW(results[i][m], float*, (nb_seqs - i - 1));

	    /* sequence 2 */
	    for(k = i + 1; k < nb_seqs; k++) {   
	       if(OPTS_first_phase < 0) { NEW(results[i][m][k - i - 1], float, 6); }
	       else { NEW(results[i][m][k - i - 1], float, 3); }
	       
	       /* phase sequence 2 */
	       for(l = OPTS_first_phase; l < 4; l++) {
		  if(l != 0) {

#ifdef SYNC_STRAND		     
		     if(((j > 0) && (l > 0)) || ((j < 0) && (l < 0))) {
#endif

		/*			INFO_ "." _INFO;*/

			n = (l > 0) ? (l - 1) : (2 - l);
			
			c = call_needlep(prots[i][m], prots[k][n], &(results[i][m][k - i - 1][n]), gapopen, gapextend, roundf(RESULTS_sequence_similarities[i][k - i - 1]));
			
			/* probleme bloquant */
			if(c < 0) return NULL;
			else {
			   
			   /* pas de pb, mais rsultat insatisfaisant */
			   if(c > 0) {
				 results[i][m][k - i - 1][n] = -INFINITY;
			   } else {
			      if(results[i][m][k - i - 1][n] < (*min)) {
				 (*min) = results[i][m][k - i - 1][n];
			      } else if(results[i][m][k - i - 1][n] > (*max)) {
				 (*max) = results[i][m][k - i - 1][n];
			      }
			   }
			}
#ifdef SYNC_STRAND
		     }
#endif
		  }
	       }
	    }
	 }
      }
   }


   INFO_ "ok\n" _INFO;

   return results;
}















void compute_pvalue_th(int nb_edges, int smin, int smax, double id, mpfr_t *pvalues, mpfr_t res) {
   double lamb = (1.0/8.0)*(id-50.0);
   mpfr_t b, expl;
   int mini, maxi;
   
   mpfr_inits(b, expl, NULL);
   
   if(nb_edges < 4) {
      mpfr_add(b, pvalues[smax-smin], pvalues[smax-smin-1], GMP_RNDN);
      mpfr_div_ui(res, b, 2, GMP_RNDN);
   } else {
      
      mpfr_set_d(b, -lamb, GMP_RNDN);
      
      mpfr_exp(expl, b, GMP_RNDN);
      mpfr_add_ui(res, expl, 1, GMP_RNDN);
      mpfr_div(b, expl, res, GMP_RNDN);

      if(OPTS_first_phase < 0){
	 maxi = (nb_edges*NB_RANK + nb_edges*(NB_RANK-1))/2;
	 /*      maxi = smax;*/
	 mini = (nb_edges*NB_RANK + nb_edges*(NB_RANK/2))/2;
      } else {
	 maxi = (nb_edges*NB_RANK + nb_edges*(NB_RANK-1))/4;
	 /*      maxi = smax;*/
	 mini = (nb_edges*NB_RANK + nb_edges*(NB_RANK/2))/4;
      }

      mpfr_div(res, pvalues[mini-smin], pvalues[maxi-smin], GMP_RNDN);
      mpfr_pow(expl, res, b, GMP_RNDN);
      mpfr_mul(res, expl, pvalues[maxi-smin], GMP_RNDN);
   }
   
   mpfr_clears(b, expl, NULL);   
}




int
do_predict_zscore(int nb_edges, double id, double bestzscore) {
   double th;

   if(OPTS_force_protea_prediction) {
      return 1; /* coding */
   }

   /*
   if(id >= 95.0) {
      th = 2.5;
   } else if(id >= 90.0) {
      th = 3.75;
   } else if(id >= 80.0) {
      th = 4.75;
   } else {
      th = 4.5;
   }
   */

   if(id >= 95.0) {
      th = 2.5;
   } else if(id >= 80.0) {
      th = -0.0077 * id * id + 1.21*id - 42.61;
   } else {
      th = 4.5;
   }

   

   if(bestzscore >= th) {
      return 1; /* coding */
   }

   return 0; /* other */
}


int
do_predict_pvalue(int nb_edges, double id, mpfr_t bestpval, mpfr_t *pvalues) {
   int smin, smax, res;
   mpfr_t pvalue_th;

   if(OPTS_force_protea_prediction) {
      return 1; /* coding */
   }


   mpfr_init(pvalue_th);

   smin = nb_edges;
   if(OPTS_first_phase < 0) {
      smax = nb_edges * NB_RANK;
   } else {
      smax = (nb_edges * NB_RANK)/2;
   }
   
   compute_pvalue_th(nb_edges, smin, smax, id, pvalues, pvalue_th);
   
   res = mpfr_cmp(bestpval, pvalue_th);

   mpfr_clear(pvalue_th);
   
   if(res > 0) {
      return 0; /* other */
   }

   return 1; /* coding */
}





int
box_needlep(const sequence* oseqs, const int onb_seqs, sequence** oprots) {
   int i, j, l, k;
   FILE* output_fd = NULL;

   const sequence* seqs = NULL;
   sequence** prots = NULL, *the_prots = NULL;
   int nb_seqs = 0;

   char* goutput = NULL;
   char* groutput = NULL;
   FILE* goutput_fd = NULL;
   FILE* groutput_fd = NULL;

   char* sw = NULL;

#ifdef PROTEA_PRE_FUSION
   char* swi = NULL;
   char* swo = NULL;
   FILE* swfd = NULL;
   alignment tdna = NULL;
   alignment taa = NULL;
   sequence* tseqs = NULL;
   int tnb_seqs;
#endif

#if defined (PROTEA_PRE_FUSION) || defined (PROTEA_POST_FUSION)
   int* equiv_class = NULL;
   int nb_class;
   sequence *dna_translated = NULL;
#else
   sequence_set cur_seq_set = NULL;
#endif

   char *output = NULL;
   int status = 0;
   float min, max;

   float gapopen, gapextend;
   int nb_edges, nb_pruned;

   float ****rfg = NULL, ****vrfg = NULL, ****rank_vrfg = NULL;

   int p;
   mpfr_t *pvalues = NULL;

   int ret = 0;

   fusion_set slinked, ssingle, sepsilon, srfg;

   list res_list = NULL;

   double avg_id;
   double pred_res;

   int *frames = NULL;
   int coding = 0;
   


   if(OPTS_output_dir) {
      if(OPTS_keep_files) {
	 NEW(sw, char, (strlen(OPTS_output_dir) + strlen(NEEDLE_PROTEIN_BOX->dir) + 1));
	 strcpy(sw, OPTS_output_dir);
	 strcat(sw, NEEDLE_PROTEIN_BOX->dir);
	 if(prepare_dir(sw)) {
	    ERROR_ "unable to create output dir '%s'\n", sw _ERROR;
	    return 1;
	 }
	 DESTROY(sw);
      }

      NEW(output, char, (strlen(OPTS_output_dir) + strlen(NEEDLE_PROTEIN_BOX->out_file) + 1));
      strcpy(output, OPTS_output_dir);
      strcat(output, NEEDLE_PROTEIN_BOX->out_file);
      if((output_fd = fopen(output, "w")) == NULL) {
	 ERROR_ "unable to create output file '%s'\n", output _ERROR;
	 return 1;
      }

      NEW(goutput, char, (strlen(OPTS_output_dir) + strlen(NEEDLE_PROTEIN_BOX->graph_out_file) + 1));
      strcpy(goutput, OPTS_output_dir);
      strcat(goutput, NEEDLE_PROTEIN_BOX->graph_out_file);
      if((goutput_fd = fopen(goutput, "w")) == NULL) {
	 ERROR_ "unable to create output file '%s'\n", goutput _ERROR;
	 return 1;
      }

      NEW(groutput, char, (strlen(OPTS_output_dir) + strlen(NEEDLE_PROTEIN_BOX->ranked_graph_out_file) + 1));
      strcpy(groutput, OPTS_output_dir);
      strcat(groutput, NEEDLE_PROTEIN_BOX->ranked_graph_out_file);
      if((groutput_fd = fopen(groutput, "w")) == NULL) {
	 ERROR_ "unable to create output file '%s'\n", groutput _ERROR;
	 return 1;
      }
   } else {
      output = NULL;
      output_fd = stdout;
      goutput = NULL;
      goutput_fd = NULL;
      groutput = NULL;
      groutput_fd = NULL;
   }


   /********** CALCUL SIMILARITE *******/

   if(RESULTS_sequence_similarities == NULL) {
      RESULTS_sequence_similarities = similarities(oseqs, onb_seqs);
   }


   if(OPTS_first_phase < 0) p = 6;
   else p = 3;



#ifdef PROTEA_REDUCE
   /********** REDUCTION **********/

   reduce_seq_set(oseqs, onb_seqs, &seqs, &nb_seqs, PROTEA_SIM_THRESHOLD);

   if(nb_seqs < 2) {
      ERROR_ "Not enough non similar sequences !!\n" _ERROR;
      fprintf(output_fd, "No graph found ...\n");

      if(output) {
	 fflush(output_fd);
	 fclose(output_fd);
	 DESTROY(output);
      }   
      
      if(goutput) {
	 fflush(goutput_fd);
	 fclose(goutput_fd);
	 DESTROY(goutput);
      }

      if(groutput) {
	 fflush(groutput_fd);
	 fclose(groutput_fd);
	 DESTROY(groutput);
      }

      return 1;
   }

   NEW(prots, sequence*, nb_seqs);
   j = 0;
   for(i = 0; i < onb_seqs; i++) {
      if(oseqs[i] == seqs[j]) {
	 prots[j] = oprots[i];
	 j++;
      }
   }

   slinked = fusion_set_new(1);
   sepsilon = fusion_set_new(1);
   ssingle = fusion_set_new(nb_seqs);
   for(i = 0; i < nb_seqs; i++) {
      cur_seq_set = sequence_set_new(1);
      sequence_set_add(cur_seq_set, seqs[i], i);
      fusion_set_add(ssingle, cur_seq_set);
   }

#else
   seqs = oseqs;
   nb_seqs = onb_seqs;


#if defined(PROTEA_PRE_FUSION) || defined (PROTEA_POST_FUSION)

   build_equiv_class(&equiv_class, &nb_class, seqs, nb_seqs, PROTEA_SIM_THRESHOLD);


   slinked = build_slinked(seqs, nb_seqs, equiv_class, nb_class);

   ssingle = build_ssingle(seqs, nb_seqs, equiv_class, nb_class);

   sepsilon = build_sepsilon(seqs, nb_seqs, slinked, ssingle);

   DESTROY(equiv_class);

#else
   slinked = fusion_set_new(1);
   sepsilon = fusion_set_new(1);
   ssingle = fusion_set_new(nb_seqs);
   for(i = 0; i < nb_seqs; i++) {
      cur_seq_set = sequence_set_new(1);
      sequence_set_add(cur_seq_set, seqs[i], i);
      fusion_set_add(ssingle, cur_seq_set);
   }

#endif





   INFO_ "Clustering sequences " _INFO;


   INFO_ "slinked={ " _INFO;
   for(i = 0; i < slinked->size; i++) {
      INFO_ "(" _INFO;
      for(j = 0; j < slinked->sets[i]->nb_seqs; j++) {
	 INFO_ "%i ", slinked->sets[i]->seq_id[j] _INFO;
      }
      INFO_ ") " _INFO;
   }
   INFO_ "} ssingle={ " _INFO;
   for(i = 0; i < ssingle->size; i++) {
      for(j = 0; j < ssingle->sets[i]->nb_seqs; j++) {
	 INFO_ "%i ", ssingle->sets[i]->seq_id[j] _INFO;
      }
   }
   INFO_ "} sepsilon={ " _INFO;
   for(i = 0; i < sepsilon->size; i++) {
      INFO_ "(" _INFO;
      for(j = 0; j < sepsilon->sets[i]->nb_seqs; j++) {
	 INFO_ "%i ", sepsilon->sets[i]->seq_id[j] _INFO;
      }
      INFO_ ") " _INFO;
   }
   INFO_ "}\n" _INFO;





#ifdef PROTEA_PRE_FUSION
   /******* construction aln multiples ADN puis deduction proteine ********/
   
   INFO_ "Translating nucleic alignments..." _INFO;
   FLUSH_OUT;


   NEW(prots, sequence*, nb_seqs);
   for(i = 0; i < nb_seqs; i++) {
      NEW(prots[i], sequence, p);
   }

   for(j = 0; j < p; j++) {
      for(i = 0; i < ssingle->size; i++) {
	 for(k = 0; k < ssingle->sets[i]->nb_seqs; k++) {
	    prots[ssingle->sets[i]->seq_id[k]][j] = init_sequence(oprots[ssingle->sets[i]->seq_id[k]][j]->file, oprots[ssingle->sets[i]->seq_id[k]][j]->input_name, oprots[ssingle->sets[i]->seq_id[k]][j]->name, oprots[ssingle->sets[i]->seq_id[k]][j]->bases, oprots[ssingle->sets[i]->seq_id[k]][j]->length, oprots[ssingle->sets[0]->seq_id[k]][j]->strand);
	 }
      }
   }

   sw = tmpnam(NULL);
   NEW(swi, char, (1+strlen(sw)));
   strcpy(swi, sw);
   sw = tmpnam(NULL);
   NEW(swo, char, (1+strlen(sw)));
   strcpy(swo, sw);

   for(i = 0; i < slinked->size; i++) {
      if((swfd = fopen(swi, "w"))) {
	 write_fasta(swfd, slinked->sets[i]->seqs, slinked->sets[i]->nb_seqs, 0);
	 fflush(swfd);
	 fclose(swfd);

	 /*	 if(OPTS_clustalw_multiple_alignment) {*/
	    if(align_clustalw(swi, swo)) {
	       ERROR_ "a problem occured during multiple alignment\n" _ERROR;
	       return 2;
	    }
	    /*} else {
	    if(align_tcoffee(swi, swo)) {
	       ERROR_ "a problem occured during multiple alignment\n" _ERROR;
	       return 2;
	    }
	    }*/

	 if(read_sequences(swo, &tseqs, &tnb_seqs)) {
	    ERROR_ "unable to open '%s' for reading...\n", swo _ERROR;
	    return 2;
	 } else {
	    tdna = init_alignment(tseqs, tnb_seqs, swo);
	    tdna->nb_seqs = tnb_seqs;
	    DESTROY(tseqs);
	    

	    for(j = OPTS_first_phase; j < 4; j++) {
	       if(j != 0) {
		  taa = translate_alignment(tdna, j, &dna_translated);
		  
		  for(k = 0; k < slinked->sets[i]->nb_seqs; k++) {
		     prots[slinked->sets[i]->seq_id[k]][(j > 0) ? (j - 1) : (2 - j)] = ungap(taa->seqs[k]);
		     
		     DESTROY(prots[slinked->sets[i]->seq_id[k]][(j > 0) ? (j - 1) : (2 - j)]->input_name);
		     NEW(prots[slinked->sets[i]->seq_id[k]][(j > 0) ? (j - 1) : (2 - j)]->input_name, char, (1+strlen(slinked->sets[i]->seqs[k]->input_name)));
		     strcpy(prots[slinked->sets[i]->seq_id[k]][(j > 0) ? (j - 1) : (2 - j)]->input_name, slinked->sets[i]->seqs[k]->input_name);
		     
		     DESTROY(taa->seqs[k]->file);
		     DESTROY(taa->seqs[k]->name);
		     DESTROY(taa->seqs[k]->input_name);
		     DESTROY(taa->seqs[k]->bases);
		  }
		  
		  DESTROY(taa->file);
		  DESTROY(taa->seqs);
		  DESTROY(taa);
	       }
	    }
	    
	    for(j = 0; j < tnb_seqs; j++) {
	       DESTROY(tdna->seqs[j]->file);
	       DESTROY(tdna->seqs[j]->name);
	       DESTROY(tdna->seqs[j]->input_name);
	       DESTROY(tdna->seqs[j]->bases);
	    }
	    DESTROY(tdna->seqs);
	    DESTROY(tdna->file);
	    DESTROY(tdna);
	 }
      } else {
	 ERROR_ "unable to open '%s' for writing...\n", swi _ERROR;
      }
   }
   INFO_ "ok\n" _INFO;

   unlink(swi);
   unlink(swo);

#else
   NEW(prots, sequence*, nb_seqs);
   for(i = 0; i < nb_seqs; i++) {
      prots[i] = oprots[i];
   }
#endif


#endif

   


/* #ifdef NB_RANK */
/*    if((nb_seqs*(nb_seqs-1))/2 >= RANKING_ABOVE) { */
   gapopen = 10.0;
   gapextend = 0.5;
/*    } else { */
/* #endif */
/*       gapopen = 7.0; */
/*       gapextend = 1.0; */
/* #ifdef NB_RANK */
/*    } */
/* #endif */
   


   /************* CONSTRUCTION RFG classique ****************/
   
   rfg = build_rfg(seqs, nb_seqs, prots, gapopen, gapextend, &min, &max);
   if(rfg == NULL) return 2;

   status = 0;

   INFO_ "Building RFG..." _INFO;
    

   /*** srfg et vrfg ***/
   srfg = fusion_set_union(slinked, ssingle);
   nb_edges = (srfg->size * (srfg->size - 1)) / 2 + sepsilon->size;
   for(i = 0; i < srfg->size; i++) {
      sequence_set_frameset(srfg->sets[i], rfg); /* assignements contraintes */
   }

   vrfg = build_vrfg(srfg, sepsilon, rfg);

   INFO_ "ok\n" _INFO;
   FLUSH_OUT;


   if(goutput) {

      write_fusion_graph(goutput_fd, vrfg, nb_seqs, seqs, srfg, sepsilon);

      fflush(goutput_fd);
      fclose(goutput_fd);
      DESTROY(goutput);
   }





   for(i = 0; i < (nb_seqs - 1); i++) {

      for(j = OPTS_first_phase; j < 4; j++) {
	 if(j != 0) {
	    k = (j > 0) ? (j - 1) : (2 - j);

	    for(l = i + 1; l < nb_seqs; l++) {	    
	       DESTROY(rfg[i][k][l-i-1]);
	    }
	    
	    DESTROY(rfg[i][k]);
	 }
      }

      DESTROY(rfg[i]);
   }

   DESTROY(rfg);




   avg_id = 0.0;

   for(i = 0; i < (nb_seqs - 1); i++) {
      for(j = i+1; j < nb_seqs; j++) {
	 avg_id += RESULTS_sequence_similarities[i][j - i - 1];
      }
   }

   avg_id = (avg_id * 2.0) / (nb_seqs*(nb_seqs - 1.0));





   NEW(frames, int, onb_seqs);


   
#ifdef NB_RANK
   if(nb_edges >= RANKING_ABOVE) {
      INFO_ "Ranking results..." _INFO;
      FLUSH_OUT;
      rank_vrfg = rank_vrfg_results(srfg->size, vrfg, srfg, sepsilon);
      INFO_ "ok\n" _INFO;


#ifdef PROTEA_PRUNING
      /********** PRUNE EDGES BETWEEN CLOSELY RELATED SEQUENCES *****/

      INFO_ "Pruning results..." _INFO;
      FLUSH_OUT;
      nb_pruned = prune_rank_vrfg(srfg->size, rank_vrfg, srfg);
      INFO_ "ok\n" _INFO;

#else
      nb_pruned = 0;
#endif


      if(groutput) {
	 
	 write_fusion_graph(groutput_fd, rank_vrfg, nb_seqs, seqs, srfg, sepsilon);
	 
	 fflush(groutput_fd);
	 fclose(groutput_fd);
	 DESTROY(groutput);
      }
      


      /** recherche de la meilleure affectation .... **/
      INFO_ "Computing and sorting best assignments..." _INFO;
      FLUSH_OUT;
      res_list = find_best_clique_fusion(srfg->size, 0.0, NB_RANK, rank_vrfg, srfg, sepsilon, 0);
      INFO_ "ok\n" _INFO;

      for(i = 0; i < (srfg->size - 1); i++) {
	 for(j = 0; j < p; j++) {
	    for(k = i + 1; k < srfg->size; k++) {
	       if((j + k - 1) > 0) {
		  DESTROY(rank_vrfg[i][j][k-i-1]);
	       }
	    }
	    if((i+j) > 0) {
	       DESTROY(rank_vrfg[i][j]);
	    }
	 }
	 if(i > 0) {
	    DESTROY(rank_vrfg[i]);
	 }
      }
      
      DESTROY(rank_vrfg);

      INFO_ "Computing pvalues " _INFO;
      INFO_ "(%s)...", "ranking system, exact" _INFO;
      FLUSH_OUT;
      pvalues = pvalues_ranking_scores_sum(nb_edges - nb_pruned);
      INFO_ "ok\n" _INFO;

      if(res_list->size > 0) {
	 fprintf(output_fd, "%i best assignments (%i pairwise alignments):\n", res_list->size, nb_edges - nb_pruned);
	 format_fusion_result_list(srfg, output_fd, res_list, seqs, nb_edges - nb_pruned, pvalues);
	 if(do_predict_pvalue(nb_edges - nb_pruned, avg_id, pvalues[-(int)res_list->elements[0].value - nb_edges - nb_pruned], pvalues)) {
	    fprintf(output_fd, "prediction = CODING\n");
	    coding = 1;

	    for(i = 0; i < onb_seqs; i++) {
	       j = 0;
	       while((j < srfg->size) && (!sequence_set_contains(srfg->sets[j], oseqs[i]))) {
		  j++;
	       }
	       frames[i] = res_list->elements[0].phases[j];
	    }

	 } else {
	    fprintf(output_fd, "prediction = OTHER\n");
	 }
      } else {
	 fprintf(output_fd, "No graph found...\n");
      }
      
      DESTROY(pvalues);
   } else {
#endif

      /** recherche de la meilleure affectation .... **/
      INFO_ "Computing and sorting best assignments..." _INFO;
      FLUSH_OUT;
      res_list = find_best_clique_fusion(srfg->size, 0.0, +INFINITY, vrfg, srfg, sepsilon, 1);
      INFO_ "ok\n" _INFO;

      if(res_list->size > 0) {
	 INFO_ "Computing zscore " _INFO;
#ifdef PROTEA_FUSION_SP
	 INFO_ "(%s)...", "full sum-of-pairs" _INFO;
#else
	 INFO_ "(%s)...", "partial sum-of-pairs" _INFO;	    
#endif
	 INFO_ "ok\n" _INFO;

	 fprintf(output_fd, "%i best assignments (%i pairwise alignments):\n", MIN(p, res_list->size), nb_edges);
	 format_fusion_result_list_zscore(srfg, output_fd, res_list, seqs, &pred_res, frames);

	 if(do_predict_zscore(nb_edges, avg_id, pred_res)) {
	    fprintf(output_fd, "prediction = CODING\n");

	    coding = 1;

	 } else {
	    fprintf(output_fd, "prediction = OTHER\n");
	 }

	 FLUSH_OUT;
      } else {
	 fprintf(output_fd, "No graph found...\n");
      }

#ifdef NB_RANK
   }
#endif

   for(i = 0; i < (srfg->size - 1); i++) {
      for(j = 0; j < p; j++) {
	 for(k = i + 1; k < srfg->size; k++) {
	    DESTROY(vrfg[i][j][k-i-1]);
	 }
	    DESTROY(vrfg[i][j]);
      }
      
      DESTROY(vrfg[i]);
   }
   
   DESTROY(vrfg);



   if((res_list->size > 0) && coding) {

      NEW(the_prots, sequence, nb_seqs);
      for(i = 0; i < nb_seqs; i++) {
	 j = 0;
	 while((j < srfg->size) && (sequence_set_contains(srfg->sets[j], seqs[i]) == 0)) {
	    j++;
	 }
	 k = 0;
	 while((k < srfg->sets[j]->nb_seqs) && (srfg->sets[j]->seqs[k] != seqs[i])) {
	    k++;
	 }

	 /*	 frames[i] = res_list->elements[0].phases[j];*/
	 the_prots[i] =  prots[i][srfg->sets[j]->frames[res_list->elements[0].phases[j]][k]];
      }

      free_list(res_list);
      
      if(compute_and_write_protein_alignment(seqs, nb_seqs, the_prots, nb_seqs, frames)) {
	 ERROR_ "unable to compute multiple alignment...\n" _ERROR;
	 ret = 1;
      }

      DESTROY(the_prots);
   }

   DESTROY(frames);

   
   if(output) {
      fflush(output_fd);
      fclose(output_fd);
      DESTROY(output);
   }

   fusion_set_destroy(srfg, 1);


#ifdef PROTEA_PRE_FUSION
   for(i = 0; i < nb_seqs; i++) {      
      for(j = 0; j < p; j++) {
	 DESTROY(prots[i][j]->file);
	 DESTROY(prots[i][j]->name);
	 DESTROY(prots[i][j]->input_name);
	 DESTROY(prots[i][j]->bases);
	 DESTROY(prots[i][j]);
      }
      DESTROY(prots[i]);
   }
#endif

   DESTROY(prots);

#ifdef PROTEA_REDUCE
   DESTROY(seqs);
#endif



   return ret;

}
