%{

#include "protea_interface.h"
#include "protea_parser.h"
#include "opts.h"
#include "display.h"
#include <string.h>
#include <assert.h>
#include <ctype.h>
%}


%option yylineno
%option nounput
%option never-interactive



SPACE           [[:blank:]]
SPACES          ({SPACE}+)
CRLF            (([\r][\n])|[\r]|[\n])
CRLFN           ({CRLF}+)

FASTA_GDE_TITLE     ([>#][^\r\n]+{CRLF})
CLUSTAL_TITLE   (CLUSTAL[^\r\n]*{CRLF})

NUMBER_ELEMENT   ([[:digit:]]+)

SEQUENCE_ELEMENT ([[:alpha:]\*\.\-]{1,10000})
CLUSTAL_SEQ_NAME ([[:alpha:][:digit:]_\~\#\-\.=+()\|/]+)
CLUSTAL_SEP_ELMT ([ \t\*.:\|]*{CRLF})


%x FASTA_GDE CLUSTAL_NAME CLUSTAL_SEQ CLUSTAL_NAME_SEP


%%

<INITIAL,FASTA_GDE>{FASTA_GDE_TITLE} {
  int nbc, i;
  i = 0;
  while(proteatext[i] == ' ' || proteatext[i] == '>') i++;
  nbc = protealeng-1;
  while((proteatext[nbc] == '\n') || (proteatext[nbc] == '\r') || (proteatext[nbc] == ' ')) nbc--;
  NEW(protealval.text, char, (1+nbc-i+1));
  strncpy(protealval.text, &(proteatext[i]), (1+nbc-i+1));
  protealval.text[nbc+1-i] = '\0';
  BEGIN(FASTA_GDE);
  if(protealval.text[0] == '>') {
    return FASTA_TITLE_T;
  }
  return GDE_TITLE_T;
}

<FASTA_GDE,CLUSTAL_SEQ>{SEQUENCE_ELEMENT} {
  int ii;

  NEW(protealval.text, char, (1 + protealeng));

  for(ii = 0; ii < protealeng; ii++) {
	 protealval.text[ii] = toupper(proteatext[ii]);
  }
  protealval.text[protealeng] = '\0';

  return SEQUENCE_ELEMENT_T;
}

<FASTA_GDE,CLUSTAL_SEQ>{NUMBER_ELEMENT} {}

<INITIAL,FASTA_GDE,CLUSTAL_NAME>{CRLFN} {}
<INITIAL,FASTA_GDE,CLUSTAL_SEQ>{SPACES} {}


<INITIAL>{CLUSTAL_TITLE} {
  BEGIN(CLUSTAL_NAME);
}


<CLUSTAL_NAME,CLUSTAL_NAME_SEP>{CLUSTAL_SEQ_NAME} {
  int nbc, i;
  i = 0;
  while(proteatext[i] == ' ') i++;
  nbc = protealeng-1;
  while((proteatext[nbc] == '\n') || (proteatext[nbc] == '\r') || (proteatext[nbc] == ' ')) nbc--;
  NEW(protealval.text, char, (1+nbc-i+1));
  strncpy(protealval.text, &(proteatext[i]), (1+nbc-i+1));
  protealval.text[nbc+1-i] = '\0';
  BEGIN(CLUSTAL_SEQ);
  return CLUSTAL_SEQ_NAME_T;
}

<CLUSTAL_NAME_SEP>{CLUSTAL_SEP_ELMT} {
  BEGIN(CLUSTAL_NAME);
  return CLUSTAL_SEP_ELMT_T;
}

<CLUSTAL_SEQ>{CRLF} { BEGIN(CLUSTAL_NAME_SEP); }

<*>. { proteaerror("invalid token"); }

%%

int proteawrap() {
  BEGIN(INITIAL);
  return 1;
}


