%{
/*
Copyright or  or CCopr. INRIA : Arnaud FONTAINE

arnaud.fontaine@inria.fr

This software is a computer program whose purpose is to predict 
significantly conserved protein coding sequences and/or RNA structure 
on a set of unaligned nucleic sequences.

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
   
#include "display.h"
#include "opts.h"
#include "protea_interface.h"
#include "alignment.h"


   char *protea_file_name;
   int protea_nb_errors;
   sequence* protea_result;

   %}


%union {
   char car;
   sequence* sequences;
   sequence sequence;
   char *text;
   protea_buffer buffer;
}


%token <text> FASTA_TITLE_T
%token <text> GDE_TITLE_T
%token <text> SEQUENCE_ELEMENT_T
%token <text> CLUSTAL_SEQ_NAME_T
%token CLUSTAL_SEP_ELMT_T


%type <buffer> sequence

%type <sequences> sequences multifasta multigde clustal_seqs clustal clustal_blocks clustal_block
%type <sequence> fasta gde clustal_seq

%start sequences

%%

sequences:
multifasta
{ protea_result = $1; }
|
multigde
{ protea_result = $1; }
|
clustal
{ protea_result = $1; }
;


multifasta:
multifasta
fasta
{
   sequence* res = $1;
   int i = 0;
   
   while(res[i]) i++;
   
   RENEW(res,sequence, (i + 2));
   res[i] = $2;
   res[i+1] = NULL;
   $$ = res;
}
|
fasta
{
   sequence* res;

   NEW(res,sequence,2);

   res[0] = $1;
   res[1] = NULL;
   $$ = res;
}
;

multigde:
multigde
gde
{
   sequence* res = $1;
   int i = 0;
   
   while(res[i]) i++;
   
   RENEW(res,sequence, (i + 2));
   res[i] = $2;
   res[i+1] = NULL;
   $$ = res;
}
|
gde
{
   sequence* res;

   NEW(res,sequence,2);

   res[0] = $1;
   res[1] = NULL;
   $$ = res;
}
;

fasta:
FASTA_TITLE_T
sequence
{
   protea_buffer seq = $2;
   sequence res;

   BUFFER_RELEASE(seq);

   res = init_sequence(protea_file_name, $1, $1, BUFFER_TOSTRING(seq), BUFFER_SIZE(seq), 1);

   DESTROY(($1));

   DESTROY_BUFFER(seq);

   $$ = res;
}
;

gde:
GDE_TITLE_T
sequence
{
   protea_buffer seq = $2;
   sequence res;

   BUFFER_RELEASE(seq);

   res = init_sequence(protea_file_name, $1, $1, BUFFER_TOSTRING(seq), BUFFER_SIZE(seq), 1);

   DESTROY(($1));

   DESTROY_BUFFER(seq);

   $$ = res;
}
;


sequence:
sequence
SEQUENCE_ELEMENT_T
{
   protea_buffer res = $1;
   register int i;
   int length;
   char *txt = $2;

   length = strlen(txt);

   if(BUFFER_MAXSIZE(res) < (BUFFER_SIZE(res) + length)) {
      BUFFER_RESIZE(res, (BUFFER_SIZE(res) + length + INPUT_BUFFER));
   }

   for(i = 0; i < length; i++) {
      BUFFER_PUT(res, txt[i]);
   }

   DESTROY(txt);

   $$ = res;
}
|
{
   protea_buffer res;

   NEW_BUFFER(res, INPUT_BUFFER);

   $$ = res;
}
;









clustal:
clustal_blocks
{ $$ = $1; }
;


clustal_blocks:
clustal_blocks
clustal_block
{
   sequence *resa = $1;
   sequence *resb = $2;

   int i,j;
   
   i = 0;
   while(resa[i]) i++;

   j = 0;
   while(resb[j]) j++;

   if(i != j) {
      proteaerror("invalid number of sequences in CLUSTALW alignment...");
   } else {

      i = 0;
      while((i < j) && (strcmp(resa[i]->name, resb[i]->name) == 0)) {
	 concat_sequence(resa[i], resb[i]);
 	 DESTROY(resb[i]->file);
 	 DESTROY(resb[i]->name);
 	 DESTROY(resb[i]->input_name);
 	 DESTROY(resb[i]->bases);
 	 DESTROY(resb[i]);
	 i++;
      }
      
      if(i < j) {
	 proteaerror("invalid sequence names in CLUSTALW alignment...");
      }

      DESTROY(resb);
   }

   $$ = resa;
}
|
clustal_block
{ $$ = $1; }
;


clustal_block:
clustal_seqs
CLUSTAL_SEP_ELMT_T
{ $$ = $1; }
;


clustal_seqs:
clustal_seqs
clustal_seq
{
   sequence* res = $1;
   int i = 0;
   
   while(res[i]) i++;
   
   RENEW(res,sequence, (i + 2));
   res[i] = $2;
   res[i+1] = NULL;
   $$ = res;
}
|
clustal_seq
{
   sequence *res = NULL;

   NEW(res,sequence,2);

   res[0] = $1;
   res[1] = NULL;
   $$ = res;
}
;

clustal_seq:
CLUSTAL_SEQ_NAME_T
SEQUENCE_ELEMENT_T
{
   sequence res = NULL;
   char* name = $1;
   char* seq = $2;

   res = init_sequence(protea_file_name, name, name, seq, strlen(seq), 1);

   DESTROY(name);
   DESTROY(seq);
   
   $$ = res;
}
;





%%


   void
      proteawarning_token(const char *s, const char *token) {
      WARN_ "'%s' line '%i' token '%s': %s\n", protea_file_name, protealineno, token, s _WARN;
   }
   
   void
      proteaerror_token(const char *s, const char *token) {
      protea_nb_errors++;
      ERROR_ "'%s' line '%i' token '%s': %s\n", protea_file_name, protealineno, token, s _ERROR;
   }
   
